/**   \file  suite_libdebug-memory.c
 *   \brief  Check Libdebug memory functions
 *  \author  François Cerbelle (Fanfan), francois@cerbelle.net
 *
 *  \internal
 *       Created:  26/06/2024
 *      Revision:  none
 * Last modified:  2024-07-26 01:33
 *      Compiler:  gcc
 *  Organization:  Cerbelle.net
 *     Copyright:  Copyright (c) 2024, François Cerbelle
 *
 *  This source code is released for free distribution under the terms of the
 *  GNU General Public License as published by the Free Software Foundation.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "debug/memtrack.h"                     /* memtrack_dumpblocks */
#include "oom.h"                                /* OOM simulation */
#include <check.h>
#include <unistd.h>                             /* pid_t getpid() */
#include <sys/stat.h>                           /* mkdir(), chdir(), mode_t */
#include <stdio.h>                              /* fprintf */
#include <stdlib.h>                             /* abort() */
#include "checktools.inc"

#include "debug/memory.h"                       /* libdebug's memory macros */

START_TEST(memory_malloc_0)
{
    char* l_result;
#ifndef NDEBUG
    size_t l_RAMSize;
    size_t l_NbBlocks;
    l_NbBlocks = memtrack_getallocatedblocks();
    l_RAMSize = memtrack_getallocatedRAM();
#endif

    l_result=NULL;
    oomtest_fill(_i,2);
    l_result = malloc(0);
    oomtest_free();
#ifndef NDEBUG
    /* Enforced by memtrack, but not glibc malloc */
    ck_assert(NULL==l_result);
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    ck_assert(0==memtrack_getblocksize(l_result));
#endif
    free(l_result);
}
END_TEST

START_TEST(memory_malloc_100)
{
    char* l_result;
#ifndef NDEBUG
    size_t l_RAMSize;
    size_t l_NbBlocks;
    l_NbBlocks = memtrack_getallocatedblocks();
    l_RAMSize = memtrack_getallocatedRAM();
#endif

    l_result=NULL;
    oomtest_fill(_i,2);
    l_result = malloc(100);
    oomtest_free();
    ck_assert((NULL!=l_result)||(oomtest_enabled()));
    if (NULL!=l_result) {
#ifndef NDEBUG
        ck_assert(l_RAMSize+100==memtrack_getallocatedRAM());
        ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
        ck_assert(100==memtrack_getblocksize(l_result));
#endif
        free(l_result);
    }
}
END_TEST

START_TEST(memory_malloc_oom)
{
    char* l_result;

#ifndef NDEBUG
    size_t l_RAMSize;
    size_t l_NbBlocks;
    l_NbBlocks = memtrack_getallocatedblocks();
    l_RAMSize = memtrack_getallocatedRAM();
#endif

    l_result=NULL;
    oomtest_fill(_i,2);
    l_result = malloc(RAMLIMIT_SOFT/2);
    oomtest_free();
    ck_assert((NULL!=l_result)||(oomtest_enabled()));
    if (NULL!=l_result) {
#ifndef NDEBUG
        ck_assert(l_RAMSize+RAMLIMIT_SOFT/2==memtrack_getallocatedRAM());
        ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
#endif
        free(l_result);
    }
#ifndef NDEBUG
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
#endif
}
END_TEST

START_TEST(memory_realloc_null_0)
{
    char* l_result;
    char* l_ptr;

#ifndef NDEBUG
    size_t l_RAMSize;
    size_t l_NbBlocks;
    l_NbBlocks = memtrack_getallocatedblocks();
    l_RAMSize = memtrack_getallocatedRAM();
#endif

    l_result=NULL;
    l_ptr=NULL;
    oomtest_fill(_i,2);
    l_result = realloc(l_ptr,0);
    oomtest_free();
    ck_assert(NULL==l_ptr);
#ifndef NDEBUG
    /* Enforced by memtrack, but not glibc malloc */
    ck_assert(NULL==l_result);
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
#endif
    if (NULL!=l_result)
        free(l_result);
}
END_TEST

START_TEST(memory_realloc_null_100)
{
    char* l_result;
    char* l_ptr;

#ifndef NDEBUG
    size_t l_RAMSize;
    size_t l_NbBlocks;
    l_NbBlocks = memtrack_getallocatedblocks();
    l_RAMSize = memtrack_getallocatedRAM();
#endif

    l_result=NULL;
    l_ptr=NULL;
    oomtest_fill(_i,2);
    l_result = realloc(l_ptr,100);
    oomtest_free();
    ck_assert((NULL!=l_result)||(oomtest_enabled()));
    ck_assert(NULL==l_ptr);
    if (NULL!=l_result) {
#ifndef NDEBUG
        ck_assert(l_RAMSize+100==memtrack_getallocatedRAM());
        ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
        ck_assert(100==memtrack_getblocksize(l_result));
#endif
        free(l_result);
    }
}
END_TEST

START_TEST(memory_realloc_null_oom)
{
    char* l_result;
    char* l_ptr;

#ifndef NDEBUG
    size_t l_RAMSize;
    size_t l_NbBlocks;
    l_NbBlocks = memtrack_getallocatedblocks();
    l_RAMSize = memtrack_getallocatedRAM();
#endif

    l_result=NULL;
    l_ptr=NULL;
    oomtest_fill(_i,2);
    l_result = realloc(l_ptr,RAMLIMIT_SOFT/2);
    oomtest_free();
    ck_assert((NULL!=l_result)||(oomtest_enabled()));
    if (NULL!=l_result) {
#ifndef NDEBUG
        ck_assert(l_RAMSize+RAMLIMIT_SOFT/2==memtrack_getallocatedRAM());
        ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
#endif
        free(l_result);
    }
    ck_assert(NULL==l_ptr);
#ifndef NDEBUG
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
#endif
}
END_TEST

#ifndef NDEBUG /* Free non-heap pointer caught by gcc */
START_TEST(memory_realloc_invalid_0)
{
    char* l_result;
    char* l_ptr;

    /* This test is supposed to trigger a SIGABRT(6) and will crash the
     * whole testsuite if not caught or not in a child process */
    forktest_only;

    l_ptr=(void*)999;
#ifdef GCOV
    if (oomtest_enabled())
        __gcov_dump();
#endif
    oomtest_fill(_i,2);
    l_result=realloc(l_ptr,0);
    /* Never reached, dead code */
    (void)l_result;
}
END_TEST
#endif

#ifndef NDEBUG /* Free non-heap pointer caught by gcc */
START_TEST(memory_realloc_invalid_100)
{
    char* l_result;
    char* l_ptr;

    /* This test is supposed to trigger a SIGABRT(6) and will crash the
     * whole testsuite if not caught or not in a child process */
    forktest_only;

    l_ptr=(void*)999;
#ifdef GCOV
    if (oomtest_enabled())
        __gcov_dump();
#endif
    oomtest_fill(_i,2);
    l_result=realloc(l_ptr,100);
    /* Never reached, dead code */
    (void)l_result;
}
END_TEST
#endif

#ifndef NDEBUG /* Free non-heap pointer caught by gcc */
START_TEST(memory_realloc_invalid_oom)
{
    void* l_result;
    char* l_ptr;

    /* This test is supposed to trigger a SIGABRT(6) and will crash the
     * whole testsuite if not caught or not in a child process */
    forktest_only;

    l_ptr=(void*)999;
#ifdef GCOV
    if (oomtest_enabled())
        __gcov_dump();
#endif
    oomtest_fill(_i,2);
    l_result=realloc(l_ptr,RAMLIMIT_SOFT*2);
    /* Never reached, dead code */
    (void)l_result;
}
END_TEST
#endif

START_TEST(memory_realloc_valid_0)
{
    void* l_result;
    char* l_ptr;
#ifndef NDEBUG
    size_t l_RAMSize;
    size_t l_NbBlocks;
    l_NbBlocks = memtrack_getallocatedblocks();
    l_RAMSize = memtrack_getallocatedRAM();
#endif

    l_ptr=malloc(100);
#ifndef NDEBUG
    ck_assert(l_RAMSize+100==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
#endif
    oomtest_fill(_i,2);
    l_result=realloc(l_ptr,0);
    oomtest_free();
    (void)l_result;
    /* Behaves as free, result can be NULL or not, depending on implementations */
#ifndef NDEBUG
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
#endif
}
END_TEST

START_TEST(memory_realloc_valid_updown)
{
    char* l_result;
    char* l_ptr;
#ifndef NDEBUG
    size_t l_RAMSize;
    size_t l_NbBlocks;
    l_NbBlocks = memtrack_getallocatedblocks();
    l_RAMSize = memtrack_getallocatedRAM();
#endif

    l_result=NULL;
    l_ptr=malloc(100);
#ifndef NDEBUG
    ck_assert(l_RAMSize+100==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
#endif
    oomtest_fill(_i,2);
    l_result=realloc(l_ptr,1000);
    oomtest_free();
    ck_assert((l_result!=NULL)||(oomtest_enabled()));
    /* Repair for next test outside of OOM context */
    if (NULL==l_result)
        l_result = realloc(l_ptr,1000);
#ifndef NDEBUG
    ck_assert(l_RAMSize+1000==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
    ck_assert(1000==memtrack_getblocksize(l_result));
#endif

    l_ptr=l_result;
    oomtest_fill(_i,2);
    l_result=realloc(l_ptr,10);
    oomtest_free();
    ck_assert((l_result!=NULL)||(oomtest_enabled()));
    /* Repair for next test outside of OOM context */
    if (NULL==l_result)
        l_result = realloc(l_ptr,10);
#ifndef NDEBUG
    ck_assert(l_RAMSize+10==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
    ck_assert(10==memtrack_getblocksize(l_result));
#endif

    free(l_result);
}
END_TEST

START_TEST(memory_realloc_valid_oom)
{
    char* l_result;
    char* l_ptr;
#ifndef NDEBUG
    size_t l_RAMSize;
    size_t l_NbBlocks;
    l_NbBlocks = memtrack_getallocatedblocks();
    l_RAMSize = memtrack_getallocatedRAM();
#endif

    l_result=NULL;
    l_ptr=malloc(100);
#ifndef NDEBUG
    ck_assert(l_RAMSize+100==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
#endif
    oomtest_fill(_i,2);
    l_result = realloc(l_ptr,RAMLIMIT_SOFT/2);
    oomtest_free();
    ck_assert((NULL!=l_result)||(oomtest_enabled()));
    if (NULL!=l_result) {
#ifndef NDEBUG
        ck_assert(l_RAMSize+RAMLIMIT_SOFT/2==memtrack_getallocatedRAM());
        ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
        ck_assert(RAMLIMIT_SOFT/2==memtrack_getblocksize(l_result));
#endif
        l_ptr=l_result;
    } else {
#ifndef NDEBUG
        ck_assert(l_RAMSize+100==memtrack_getallocatedRAM());
        ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
        ck_assert(100==memtrack_getblocksize(l_ptr));
#endif
    }
    ck_assert(NULL!=l_ptr);

    /* CppCheck detects a double free here for l_ptr, first freed at realloc.
     * But if realloc fails at runtime (this is wanted), l_ptr is left
     * untouched, thus not freed */
    free(l_ptr);
    /* CppCheck detects a memory leak here for l_result, allocated at realloc.
     * But if realloc fails at runtime (which is wanted here), l_result is NULL,
     * confirmed by the assertion, and do not need to be freed */
}
END_TEST

START_TEST(memory_calloc_0_0)
{
    char* l_result;
#ifndef NDEBUG
    size_t l_RAMSize;
    size_t l_NbBlocks;
    l_NbBlocks = memtrack_getallocatedblocks();
    l_RAMSize = memtrack_getallocatedRAM();
#endif

    l_result=NULL;
    oomtest_fill(_i,2);
    l_result = calloc(0,0);
    oomtest_free();
#ifndef NDEBUG
    /* Enforced by memtrack, but not glibc malloc */
    ck_assert(NULL==l_result);
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    ck_assert(0==memtrack_getblocksize(l_result));
#endif

    free(l_result);
}
END_TEST

START_TEST(memory_calloc_0_100)
{
    char* l_result;
#ifndef NDEBUG
    size_t l_RAMSize;
    size_t l_NbBlocks;
    l_NbBlocks = memtrack_getallocatedblocks();
    l_RAMSize = memtrack_getallocatedRAM();
#endif

    l_result=NULL;
    oomtest_fill(_i,2);
    l_result = calloc(0,100);
    oomtest_free();
#ifndef NDEBUG
    /* Enforced by memtrack, but not glibc malloc */
    ck_assert(NULL==l_result);
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    ck_assert(0==memtrack_getblocksize(l_result));
#endif

    free(l_result);
}
END_TEST

START_TEST(memory_calloc_100_0)
{
    char* l_result;
#ifndef NDEBUG
    size_t l_RAMSize;
    size_t l_NbBlocks;
    l_NbBlocks = memtrack_getallocatedblocks();
    l_RAMSize = memtrack_getallocatedRAM();
#endif

    l_result=NULL;
    oomtest_fill(_i,2);
    l_result = calloc(100,0);
    oomtest_free();
#ifndef NDEBUG
    /* Enforced by memtrack, but not glibc malloc */
    ck_assert(NULL==l_result);
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    ck_assert(0==memtrack_getblocksize(l_result));
#endif

    free(l_result);
}
END_TEST

START_TEST(memory_calloc_100_100)
{
    char* l_result;
#ifndef NDEBUG
    size_t l_RAMSize;
    size_t l_NbBlocks;
    l_NbBlocks = memtrack_getallocatedblocks();
    l_RAMSize = memtrack_getallocatedRAM();
#endif

    l_result=NULL;
    oomtest_fill(_i,2);
    l_result = calloc(100,100);
    oomtest_free();
    ck_assert((NULL!=l_result)||(oomtest_enabled()));
    if (NULL!=l_result) {
#ifndef NDEBUG
        ck_assert(l_RAMSize+10000==memtrack_getallocatedRAM());
        ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
        ck_assert(10000==memtrack_getblocksize(l_result));
#endif

        free(l_result);
    }
}
END_TEST

START_TEST(memory_calloc_oom)
{
    char* l_result;
#ifndef NDEBUG
    size_t l_RAMSize;
    size_t l_NbBlocks;
    l_NbBlocks = memtrack_getallocatedblocks();
    l_RAMSize = memtrack_getallocatedRAM();
#endif

    l_result=NULL;
    oomtest_fill(_i,2);
    l_result = calloc(RAMLIMIT_SOFT/2/1024,1024);
    oomtest_free();
    ck_assert((NULL!=l_result)||(oomtest_enabled()));
    if (NULL!=l_result) {
#ifndef NDEBUG
        ck_assert(l_RAMSize+RAMLIMIT_SOFT/2==memtrack_getallocatedRAM());
        ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
#endif
        free(l_result);
    }
#ifndef NDEBUG
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
#endif
}
END_TEST

START_TEST(memory_free_null)
{
    char* l_ptr;
#ifndef NDEBUG
    size_t l_RAMSize;
    size_t l_NbBlocks;
    l_NbBlocks = memtrack_getallocatedblocks();
    l_RAMSize = memtrack_getallocatedRAM();
#endif

    l_ptr=NULL;
    oomtest_fill(_i,2);
    free(l_ptr);
    oomtest_free();
    ck_assert(NULL==l_ptr);
#ifndef NDEBUG
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
#endif
}
END_TEST

#ifndef NDEBUG /* Free non-heap pointer caught by gcc */
START_TEST(memory_free_invalid)
{
    char* l_ptr;

    l_ptr=(void*)999;
    /* This test is supposed to trigger a SIGABRT(6) and will crash the
     * whole testsuite if not caught or not in a child process */
    forktest_only;

#ifdef GCOV
    if (oomtest_enabled())
        __gcov_dump();
#endif
    oomtest_fill(_i,2);
    free(l_ptr);
}
END_TEST
#endif

START_TEST(memory_free_valid)
{
    char* l_ptr;
#ifndef NDEBUG
    size_t l_RAMSize;
    size_t l_NbBlocks;
    l_NbBlocks = memtrack_getallocatedblocks();
    l_RAMSize = memtrack_getallocatedRAM();
#endif

    l_ptr=malloc(100);
#ifndef NDEBUG
    ck_assert(l_RAMSize+100==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
#endif
    oomtest_fill(_i,2);
    free(l_ptr);
    oomtest_free();
    ck_assert(NULL!=l_ptr);
#ifndef NDEBUG
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
#endif
}
END_TEST

#ifndef NDEBUG /* strdup(NULL) caught by gcc */
START_TEST(memory_strdup_null)
{
    void* l_result;
    size_t l_RAMSize;
    size_t l_NbBlocks;

    l_NbBlocks = memtrack_getallocatedblocks();
    l_RAMSize = memtrack_getallocatedRAM();

    /* This test is supposed to trigger a SIGABRT(6) and will crash the
     * whole testsuite if not caught or not in a child process */
    forktest_only;

    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    oomtest_fill(_i,2);
    /* CppCheck warns about null pointer dereference, normal for this test */
    l_result=strdup(NULL); /* cppcheck-suppress nullPointer */
    /* Never reached, dead code */
    (void)l_result;
}
END_TEST
#endif

START_TEST(memory_strdup_empty)
{
    char* l_result;
#ifndef NDEBUG
    size_t l_RAMSize;
    size_t l_NbBlocks;
    l_NbBlocks = memtrack_getallocatedblocks();
    l_RAMSize = memtrack_getallocatedRAM();
#endif

    l_result=NULL;
#ifndef NDEBUG
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
#endif
    oomtest_fill(_i,2);
    l_result=strdup("");
    oomtest_free();
    ck_assert((NULL!=l_result)||(oomtest_enabled()));
    if (NULL!=l_result) {
#ifndef NDEBUG
        ck_assert(l_RAMSize+1==memtrack_getallocatedRAM());
        ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
#endif
        free(l_result);
    } else {
#ifndef NDEBUG
        ck_assert(l_RAMSize==memtrack_getallocatedRAM());
        ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
#endif
    }
}
END_TEST

START_TEST(memory_strdup_value)
{
    char *szLongString =
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        ;
    char* l_result;
#ifndef NDEBUG
    size_t l_RAMSize;
    size_t l_NbBlocks;
    l_NbBlocks = memtrack_getallocatedblocks();
    l_RAMSize = memtrack_getallocatedRAM();
#endif

    l_result=NULL;
#ifndef NDEBUG
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
#endif
    oomtest_fill(_i,2);
    l_result=strdup(szLongString);
    oomtest_free();
    ck_assert((NULL!=l_result)||(oomtest_enabled()));
    if (NULL!=l_result) {
#ifndef NDEBUG
        ck_assert(l_RAMSize+strlen(szLongString)+1==memtrack_getallocatedRAM());
        ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
#endif
        free(l_result);
    } else {
#ifndef NDEBUG
        ck_assert(l_RAMSize==memtrack_getallocatedRAM());
        ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
#endif
    }
}
END_TEST

START_TEST(memory_strdup_invalid)
{
    void* l_result;
#ifndef NDEBUG
    size_t l_RAMSize;
    size_t l_NbBlocks;
    l_NbBlocks = memtrack_getallocatedblocks();
    l_RAMSize = memtrack_getallocatedRAM();
#endif

    /* This test is supposed to trigger a SIGABRT(6) and will crash the
     * whole testsuite if not caught or not in a child process */
    forktest_only;

#ifndef NDEBUG
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
#endif
    oomtest_fill(_i,2);
    l_result=strdup((void*)999);
    /* Never reached, dead code */
    (void)l_result;
}
END_TEST

START_TEST(memory_strdup_oom)
{
    char *szLongString;
    size_t szSize;
    char* l_result;
#ifndef NDEBUG
    size_t l_RAMSize;
    size_t l_NbBlocks;
    l_NbBlocks = memtrack_getallocatedblocks();
    l_RAMSize = memtrack_getallocatedRAM();
#endif

    l_result=NULL;
    szSize=RAMLIMIT_HARD;
    /* Find szSize value such as malloc(szSize) works, malloc(2*szSize) fails */
    while ((szSize>0)&&(NULL==(szLongString=malloc(szSize))))
        szSize /= 2;
    szLongString = memset(szLongString,'A',szSize-1);
    szLongString[szSize-1]=0;
#ifndef NDEBUG
    ck_assert(l_RAMSize+szSize==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
#endif
    oomtest_fill(_i,2);
    l_result=strdup(szLongString);
    oomtest_free();
    ck_assert(NULL==l_result);
#ifndef NDEBUG
    ck_assert(l_RAMSize+szSize==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
#endif
    free(szLongString);
}
END_TEST

START_TEST(memory_asprintf)
{
    int l_result;
    char* l_ptr;
    char *szLongString =
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        ;
#ifndef NDEBUG
    size_t l_RAMSize;
    size_t l_NbBlocks;
    l_NbBlocks = memtrack_getallocatedblocks();
    l_RAMSize = memtrack_getallocatedRAM();
#endif

    /* String empty */
    l_result=0;
    l_ptr = NULL;
#ifndef NDEBUG
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
#endif

    oomtest_fill(_i,2);
    l_result = asprintf(&l_ptr,"%s","");
    oomtest_free();
    ck_assert((0==l_result)||(oomtest_enabled()));
    if (0==l_result) {
        ck_assert(NULL!=l_ptr);
        ck_assert(0==strcmp("",l_ptr));
#ifndef NDEBUG
        ck_assert(l_RAMSize+1==memtrack_getallocatedRAM());
        ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
        ck_assert(1==memtrack_getblocksize(l_ptr));
#endif
        free(l_ptr);
    }

    /* String value */
    l_result=0;
    l_ptr = NULL;
#ifndef NDEBUG
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
#endif

    oomtest_fill(_i,2);
    l_result = asprintf(&l_ptr,"%s",szLongString);
    oomtest_free();
    ck_assert((strlen(szLongString)==(size_t)l_result)||(oomtest_enabled()));
    if (strlen(szLongString)==(size_t)l_result) {
        ck_assert(NULL!=l_ptr);
        ck_assert(0==strcmp(szLongString,l_ptr));
#ifndef NDEBUG
        ck_assert(l_RAMSize+1+strlen(szLongString)==memtrack_getallocatedRAM());
        ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
        ck_assert(1+strlen(szLongString)==memtrack_getblocksize(l_ptr));
#endif
        free(l_ptr);
    }

    /* Numeric value */
    l_result=0;
    l_ptr = NULL;
#ifndef NDEBUG
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
#endif

    oomtest_fill(_i,2);
    l_result = asprintf(&l_ptr,"%d",123);
    oomtest_free();
    ck_assert((3==l_result)||(oomtest_enabled()));
    if (3==l_result) {
        ck_assert(NULL!=l_ptr);
        ck_assert(0==strcmp("123",l_ptr));
#ifndef NDEBUG
        ck_assert(l_RAMSize+4==memtrack_getallocatedRAM());
        ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
        ck_assert(4==memtrack_getblocksize(l_ptr));
#endif
        free(l_ptr);
    }
}
END_TEST

START_TEST(memory_memreport)
{
    oomtest_fill(_i,2);
    memreport();
    oomtest_free();
}
END_TEST

/************************************************************************/
void memory_checked_uninitialized_setup()
{
    signals_catch();
    forktest_gprofdir();
}
void memory_checked_uninitialized_teardown()
{
    signals_release();
#ifndef NDEBUG
    ck_assert(0==memtrack_getallocatedblocks());
#endif /* NDEBUG */
}

#define DATASIZE 5
#define DATACOUNT 5
char* data[DATACOUNT];
void memory_checked_withdata_setup()
{
    unsigned int i;
    signals_catch();
    forktest_gprofdir();
    for (i=0; i<DATACOUNT; i++)
        data[i]=malloc(DATASIZE);
}
void memory_checked_withdata_teardown()
{
    unsigned int i;
    signals_release();
    for (i=0; i<DATACOUNT; i++)
        free(data[i]);
}
/************************************************************************/
void memory_unchecked_common_setup()
{
    forktest_init();
}
void memory_unchecked_common_teardown()
{
}
void memory_unchecked_oom_setup()
{
    oomtest_enable(RAMLIMIT_SOFT);
    memory_unchecked_common_setup();
}
void memory_unchecked_oom_teardown()
{
    oomtest_disable();
    memory_unchecked_common_teardown();
}
/************************************************************************/
Suite* libdebug_memory_suite()
{
    Suite *s;
    TCase *tc;

    s = suite_create("LibDebug-Memory");
    tc = tcase_create("Uninitialized");
    tcase_add_checked_fixture(tc, memory_checked_uninitialized_setup, memory_checked_uninitialized_teardown);
    tcase_add_unchecked_fixture(tc, memory_unchecked_common_setup, memory_unchecked_common_teardown);
    suite_add_tcase(s, tc);
    tcase_add_test(tc, memory_malloc_0);
    tcase_add_test(tc, memory_malloc_100);
    tcase_add_test(tc, memory_malloc_oom);
    tcase_add_test(tc, memory_realloc_null_0);
    tcase_add_test(tc, memory_realloc_null_100);
    tcase_add_test(tc, memory_realloc_null_oom);
#ifndef NDEBUG /* Free non-heap pointer caught by gcc */
    tcase_add_test_raise_signal(tc, memory_realloc_invalid_0,6);
    tcase_add_test_raise_signal(tc, memory_realloc_invalid_100,6);
    tcase_add_test_raise_signal(tc, memory_realloc_invalid_oom,6);
#endif
    tcase_add_test(tc, memory_realloc_valid_0);
    tcase_add_test(tc, memory_realloc_valid_updown);
    tcase_add_test(tc, memory_realloc_valid_oom);
    tcase_add_test(tc, memory_calloc_0_0);
    tcase_add_test(tc, memory_calloc_0_100);
    tcase_add_test(tc, memory_calloc_100_0);
    tcase_add_test(tc, memory_calloc_100_100);
    tcase_add_test(tc, memory_calloc_oom);
    tcase_add_test(tc, memory_free_null);
#ifndef NDEBUG /* Free non-heap pointer caught by gcc */
    tcase_add_test_raise_signal(tc, memory_free_invalid,6);
#endif
    tcase_add_test(tc, memory_free_valid);
#ifndef NDEBUG /* strdup(NULL) caught by gcc */
    tcase_add_test_raise_signal(tc, memory_strdup_null,11);
#endif
    tcase_add_test(tc, memory_strdup_empty);
    tcase_add_test(tc, memory_strdup_value);
    tcase_add_test_raise_signal(tc, memory_strdup_invalid,11);
    tcase_add_test(tc, memory_strdup_oom);
    tcase_add_test(tc, memory_asprintf);
    tcase_add_test(tc, memory_memreport);

    tc = tcase_create("WithData");
    tcase_add_checked_fixture(tc, memory_checked_withdata_setup, memory_checked_withdata_teardown);
    tcase_add_unchecked_fixture(tc, memory_unchecked_common_setup, memory_unchecked_common_teardown);
    suite_add_tcase(s, tc);
    tcase_add_test(tc, memory_malloc_0);
    tcase_add_test(tc, memory_malloc_100);
    tcase_add_test(tc, memory_malloc_oom);
    tcase_add_test(tc, memory_realloc_null_0);
    tcase_add_test(tc, memory_realloc_null_100);
    tcase_add_test(tc, memory_realloc_null_oom);
#ifndef NDEBUG /* Free non-heap pointer caught by gcc */
    tcase_add_test_raise_signal(tc, memory_realloc_invalid_0,6);
    tcase_add_test_raise_signal(tc, memory_realloc_invalid_100,6);
    tcase_add_test_raise_signal(tc, memory_realloc_invalid_oom,6);
#endif
    tcase_add_test(tc, memory_realloc_valid_0);
    tcase_add_test(tc, memory_realloc_valid_updown);
    tcase_add_test(tc, memory_realloc_valid_oom);
    tcase_add_test(tc, memory_calloc_0_0);
    tcase_add_test(tc, memory_calloc_0_100);
    tcase_add_test(tc, memory_calloc_100_0);
    tcase_add_test(tc, memory_calloc_100_100);
    tcase_add_test(tc, memory_calloc_oom);
    tcase_add_test(tc, memory_free_null);
#ifndef NDEBUG /* Free non-heap pointer caught by gcc */
    tcase_add_test_raise_signal(tc, memory_free_invalid,6);
#endif
    tcase_add_test(tc, memory_free_valid);
#ifndef NDEBUG /* strdup(NULL) caught by gcc */
    tcase_add_test_raise_signal(tc, memory_strdup_null,11);
#endif
    tcase_add_test(tc, memory_strdup_empty);
    tcase_add_test(tc, memory_strdup_value);
    tcase_add_test_raise_signal(tc, memory_strdup_invalid,11);
    tcase_add_test(tc, memory_strdup_oom);
    tcase_add_test(tc, memory_asprintf);
    tcase_add_test(tc, memory_memreport);

    tc = tcase_create("Uninitialized-OOM");
    tcase_set_tags(tc,"oom");
    tcase_add_checked_fixture(tc, memory_checked_uninitialized_setup, memory_checked_uninitialized_teardown);
    tcase_add_unchecked_fixture(tc, memory_unchecked_oom_setup, memory_unchecked_oom_teardown);
    suite_add_tcase(s, tc);
    tcase_add_loop_test(tc, memory_malloc_0, 0, 10);
    tcase_add_loop_test(tc, memory_malloc_100, 0, 10);
    tcase_add_loop_test(tc, memory_malloc_oom, 0, 10);
    tcase_add_loop_test(tc, memory_realloc_null_0, 0, 10);
    tcase_add_loop_test(tc, memory_realloc_null_100, 0, 10);
    tcase_add_loop_test(tc, memory_realloc_null_oom, 0, 10);
#ifndef NDEBUG /* Free non-heap pointer caught by gcc */
    tcase_add_loop_test_raise_signal(tc, memory_realloc_invalid_0,6, 0, 10);
    tcase_add_loop_test_raise_signal(tc, memory_realloc_invalid_100,6, 0, 10);
    tcase_add_loop_test_raise_signal(tc, memory_realloc_invalid_oom,6, 0, 10);
#endif
    tcase_add_loop_test(tc, memory_realloc_valid_0, 0, 10);
    tcase_add_loop_test(tc, memory_realloc_valid_updown, 0, 10);
    tcase_add_loop_test(tc, memory_realloc_valid_oom, 0, 10);
    tcase_add_loop_test(tc, memory_calloc_0_0, 0, 10);
    tcase_add_loop_test(tc, memory_calloc_0_100, 0, 10);
    tcase_add_loop_test(tc, memory_calloc_100_0, 0, 10);
    tcase_add_loop_test(tc, memory_calloc_100_100, 0, 10);
    tcase_add_loop_test(tc, memory_calloc_oom, 0, 10);
    tcase_add_loop_test(tc, memory_free_null, 0, 10);
#ifndef NDEBUG /* Free non-heap pointer caught by gcc */
    tcase_add_loop_test_raise_signal(tc, memory_free_invalid,6, 0, 10);
#endif
    tcase_add_loop_test(tc, memory_free_valid, 0, 10);
#ifndef NDEBUG /* strdup(NULL) caught by gcc */
    tcase_add_loop_test_raise_signal(tc, memory_strdup_null,11, 0, 10);
#endif
    tcase_add_loop_test(tc, memory_strdup_empty, 0, 10);
    tcase_add_loop_test(tc, memory_strdup_value, 0, 10);
    tcase_add_loop_test_raise_signal(tc, memory_strdup_invalid,11, 0, 10);
    tcase_add_loop_test(tc, memory_strdup_oom, 0, 10);
    tcase_add_loop_test(tc, memory_asprintf, 0, 10);
    tcase_add_loop_test(tc, memory_memreport, 0, 10);

    tc = tcase_create("WithData-OOM");
    tcase_set_tags(tc,"oom");
    tcase_add_checked_fixture(tc, memory_checked_withdata_setup, memory_checked_withdata_teardown);
    tcase_add_unchecked_fixture(tc, memory_unchecked_oom_setup, memory_unchecked_oom_teardown);
    suite_add_tcase(s, tc);
    tcase_add_loop_test(tc, memory_malloc_0, 0, 10);
    tcase_add_loop_test(tc, memory_malloc_100, 0, 10);
    tcase_add_loop_test(tc, memory_malloc_oom, 0, 10);
    tcase_add_loop_test(tc, memory_realloc_null_0, 0, 10);
    tcase_add_loop_test(tc, memory_realloc_null_100, 0, 10);
    tcase_add_loop_test(tc, memory_realloc_null_oom, 0, 10);
#ifndef NDEBUG /* Free non-heap pointer caught by gcc */
    tcase_add_loop_test_raise_signal(tc, memory_realloc_invalid_0,6, 0, 10);
    tcase_add_loop_test_raise_signal(tc, memory_realloc_invalid_100,6, 0, 10);
    tcase_add_loop_test_raise_signal(tc, memory_realloc_invalid_oom,6, 0, 10);
#endif
    tcase_add_loop_test(tc, memory_realloc_valid_0, 0, 10);
    tcase_add_loop_test(tc, memory_realloc_valid_updown, 0, 10);
    tcase_add_loop_test(tc, memory_realloc_valid_oom, 0, 10);
    tcase_add_loop_test(tc, memory_calloc_0_0, 0, 10);
    tcase_add_loop_test(tc, memory_calloc_0_100, 0, 10);
    tcase_add_loop_test(tc, memory_calloc_100_0, 0, 10);
    tcase_add_loop_test(tc, memory_calloc_100_100, 0, 10);
    tcase_add_loop_test(tc, memory_calloc_oom, 0, 10);
    tcase_add_loop_test(tc, memory_free_null, 0, 10);
#ifndef NDEBUG /* Free non-heap pointer caught by gcc */
    tcase_add_loop_test_raise_signal(tc, memory_free_invalid,6, 0, 10);
#endif
    tcase_add_loop_test(tc, memory_free_valid, 0, 10);
#ifndef NDEBUG /* strdup(NULL) caught by gcc */
    tcase_add_loop_test_raise_signal(tc, memory_strdup_null,11, 0, 10);
#endif
    tcase_add_loop_test(tc, memory_strdup_empty, 0, 10);
    tcase_add_loop_test(tc, memory_strdup_value, 0, 10);
    tcase_add_loop_test_raise_signal(tc, memory_strdup_invalid,11, 0, 10);
    tcase_add_loop_test(tc, memory_strdup_oom, 0, 10);
    tcase_add_loop_test(tc, memory_asprintf, 0, 10);
    tcase_add_loop_test(tc, memory_memreport, 0, 10);

    return s;
}

